#include "util.h"
#include "waveform_display.h"

static int half_rand = RAND_MAX/2;

int Waveform_Display::setpoint(int which, int value) {
  int lastvalue = points[which]->y();
  int where = which*visual_data_ratio;
  points[which]->position(points[which]->x(),
			  clip_range(value, point_start_y, point_end_y));
  for(int i = where; i < where+visual_data_ratio; i++) {
    audio_data[i] = ((points[which]->y()*data_scale)+data_offset)*-1;
  }
  return lastvalue;
}

Waveform_Display::Waveform_Display(int x, int y, int width, int height,
				   int count, int *l, int *r, unsigned ratio)
  : Fl_Group(x, y, width, height) {
  left = l;
  right = r;
  scatter = 0;
  point_size = (float)width/(float)count;
  point_size = (point_size < 1 ) ? 1 : point_size;
  point_count = count;
  point_start_x = x;
  point_start_y = y+1;
  point_end_y = y+height- (int) point_size;
  points = new Fl_Box*[count];
  range = point_end_y-point_start_y;
  middle = (range/2)+point_start_y;
  box(FL_PLASTIC_DOWN_BOX);
  color(FL_FOREGROUND_COLOR);
  if (paste_buffer_size < count) {
    paste_buffer = (int *) realloc (paste_buffer, count*sizeof(int));
    paste_buffer_size = count;
  }
  for (int i = 0; i < count; i++) {
    points[i]= new Fl_Box(point_start_x+(int)(i*point_size),
			  middle, (int) (point_size+.4), (int) (point_size+.4));
    points[i]->box(FL_PLASTIC_UP_BOX);
    points[i]->color(FL_FOREGROUND_COLOR);
  }
  end();
  visual_data_ratio = ratio;
  data_count = count*ratio;
  audio_data = (float *) malloc (data_count*sizeof(float));
  data_scale = (float) 2 / (float) range;
  data_offset = -1.0f - point_start_y*data_scale;
}

int Waveform_Display::x_to_point(int _x) {
  return clip_range(int((_x-x())/point_size), 0, point_count-1);
}

int Waveform_Display::handle(int event) {
  static int last_point, last_value;
  static int last_x, last_y;
  int which_point, value;
  switch (event) {
  case FL_PUSH:
  case FL_RELEASE:
    last_point = -1;
  case FL_DRAG:
    which_point = clip_range(x_to_point(Fl::event_x()),
			     x_to_point(*left),
			     x_to_point(*right-(int)point_size/2));
    value = Fl::event_y()-(int)(point_size/2);
    if ((last_point > -1) && !scatter && last_point != which_point) {
      int start_point, end_point, increment, index_value;
      if (last_point < which_point) {
	start_point = last_point;
	end_point = which_point;
	increment = (last_value - value)/(start_point-end_point);
	index_value = last_value;
      } else {
	start_point = which_point;
	end_point = last_point;
	increment = (value - last_value)/(start_point-end_point);
	index_value = value;
      }
      for(int i = start_point; i < end_point; i++, index_value += increment) {
	setpoint(i, index_value);
      }
      clip(Fl::event_x(), last_x, Fl::event_y(), last_y,
	   (int)point_size, (int)point_size);
      redraw();
      fl_pop_clip();
    } else {
      setpoint(which_point, value);
      fl_push_clip(Fl::event_x(), Fl::event_y(), (int)point_size, (int)point_size);
      redraw();
      fl_pop_clip();
    }
    last_point = which_point;
    last_value = value;
    return 1;
  }
  return 0;
}

void Waveform_Display::manipulate(enum actions how) {
  int left_edge =  x_to_point(*left+(int)point_size);
  int right_edge = x_to_point(*right-(int)(point_size/2))+1;

  switch(how) {
  case rndm:
    for(unsigned i = left_edge; i < right_edge; i++) {
      setpoint(i, (int)(range*(float)rand()/RAND_MAX)+point_start_y);
    }
    break;
  case drunk:
    {
      float velocity, spread = (float) range/10.0;
      for(unsigned i = left_edge; i < right_edge; i++) {
	velocity = ((float) (rand() - half_rand)) / RAND_MAX;
	setpoint(i, points[i]->y()+ (int) (spread*velocity));
      }
    }
    break;
  case fill:
    {
      int min = point_end_y;
      int max = point_start_y;
      int offset;
      float scale;
      for(unsigned i = left_edge; i < right_edge; i++) {
	min = (points[i]->y() < min) ? points[i]->y() : min;
	max = (points[i]->y() > max) ? points[i]->y() : max;
      }
      if(max != point_end_y || min != point_start_y) {
	scale = (float) range / (float) (max - min);
	for(unsigned i = left_edge; i < right_edge; i++) {
	  if(points[i]->y() != middle) {
	    int distance = points[i]->y()-min;
	    setpoint(i, (int)(scale*distance)+point_start_y);
	  }
	}
      }
    }
    break;
  case zero:
    for(unsigned i = left_edge; i < right_edge; i++) {
      setpoint(i, middle);
    }
    break;
  case grow:
    for(unsigned i = left_edge; i < right_edge; i++) {
      setpoint(i, (int)(((float)(points[i]->y()-middle)*
			 1.11111111111111112)
			+(float)middle));
    }
    break;
  case shrink:
    for(unsigned i = left_edge; i < right_edge; i++) {
      setpoint(i, (int)((float)(points[i]->y()-middle)*.9+(float)middle));
    }
    break;
  case copy:
    paste_buffer_used = right_edge - left_edge;
    paste_buffer_range = range;
    for(unsigned i = 0; i < paste_buffer_used; i++) {
      paste_buffer[i] = points[left_edge+i]->y()-point_start_y;
    }
    break;
  case paste:
    for(unsigned i = left_edge, buf_i = 0;
	i < right_edge && buf_i < paste_buffer_used;
	i++, buf_i++) {
      setpoint(i, (int) (paste_buffer[buf_i] *
			 ((float)range/(float)paste_buffer_range))
	       +point_start_y);
    }
    break;
  case smooth: {
    int min = point_end_y;
    int new_min = point_end_y;
    int max = point_start_y;
    int new_max = point_start_y;
    int offset;
    float scale;
    float collector = points[right_edge-1]->y();

    for (unsigned i = left_edge; i < right_edge; i++) {
      min = (points[i]->y() < min) ? points[i]->y() : min;
      max = (points[i]->y() > max) ? points[i]->y() : max;
      collector += (float)(points[i]->y());
      collector /= 2.f;
      setpoint(i, (int) collector);
      new_min = (points[i]->y() < new_min) ? points[i]->y() : new_min;
      new_max = (points[i]->y() > new_max) ? points[i]->y() : new_max;
    }
    scale = (max - min) / (new_max - new_min);
    offset = min - new_min;
    for(unsigned i = left_edge; i < right_edge; i++) {
      setpoint(i, (int) (offset + point_start_y +
			 (points[i]->y()-point_start_y)*scale));
    }
  }
    break;
  case rough: {
    float collector = points[right_edge-1]->y();
    for (unsigned i = left_edge+1; i < right_edge; i++) {
      collector -= (float) middle;
      collector *= -1.f;
      collector /= 4.f;
      collector += (float) middle;
      collector += (float)points[i]->y();
      collector /= 2.f;
      setpoint(i, (int)collector);
      collector = (float) points[i]->y();
    }
  }
    break;
  case up:
    for (unsigned i = left_edge; i < right_edge; i++) {
      setpoint(i, points[i]->y()-5);
    }
    break;
  case down:
    for (unsigned i = left_edge; i < right_edge; i++) {
      setpoint(i, points[i]->y()+5);
    }
  }
  redraw();
}
